// =======================================================================
// This file is part of sswriter (Sky Status Writer).
//
// Copyright 2008  Jonathan 'Sky' Squirawski <webmaster@sky-siteweb.com>
// 
// SSWriter is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// SSWriter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SSWriter.  If not, see <http://www.gnu.org/licenses/>.
// =======================================================================

#include "acpi_parsing.h"

char buff[BUFF_SZ];
char path[PATH_SZ];

// -----------------------------------------------------------------------
//   Append the current date.
// -----------------------------------------------------------------------
unsigned int insert_date(char *sbuff, unsigned int len,
      const char *date_format) {
   time_t t;
   struct tm *t_tmp;

   // Setting locale
   setlocale(LC_TIME, "");
   time(&t);
   t_tmp = localtime(&t);

   if (date_format != NULL) {
      return strftime(sbuff, len, date_format, t_tmp);
   } else {
      return strftime(sbuff, len, "%c", t_tmp);
   }
}

// -----------------------------------------------------------------------
//   Append the state of the battery.
//   Put in warn if the warn_limit is passed.
// -----------------------------------------------------------------------
unsigned int insert_battery_state(char *sbuff, unsigned int len,
      unsigned int number, unsigned int warn_limit, int *warn) {
   FILE *f;
   int bat_rem_capacity = -1;
   int bat_des_capacity = -1;
   int bat_percent = -1;
   char bat_state[32] = "";

   snprintf(path, PATH_SZ, "%s%d%s", BAT, number, BAT_STATE);

   f = fopen(path, "r");
   if (f == NULL) return 0;

   int found = 0;
   while (fgets(buff, BUFF_SZ, f)) {
      if (bat_rem_capacity < 0
            && sscanf(buff, "remaining capacity:\t%d", &bat_rem_capacity))
         continue;
      else if (!found && sscanf(buff, "charging state:\t%s", bat_state)) {
         found = 1;
         continue;
      }
   }
   fclose(f);

   // If unable to read, return
   if (bat_rem_capacity < 0 || bat_state[0] == '\0')
      return 0;

   snprintf(path, PATH_SZ, "%s%d%s", BAT, number, BAT_INFO);

   f = fopen(path, "r");
   if (f == NULL) return 0;

   while (fgets(buff, BUFF_SZ, f)) {
      if (sscanf(buff, "last full capacity:\t%d", &bat_des_capacity))
         break;
   }
   fclose(f);

   if (bat_des_capacity <= 0) return 0;

   bat_percent = (bat_rem_capacity * 100) / bat_des_capacity;

   // Check the level of battery
   if (strncmp(bat_state, "discharging", 32) == 0
         && bat_percent <= warn_limit) { 
      *warn = 1;
      return snprintf(sbuff, len, "# WARNING BATTERY IS LOW # %s %d%%",
            bat_state, bat_percent);
   }

   *warn = 0;
   return snprintf(sbuff, len, "%s %d%%", bat_state, bat_percent);
}

// -----------------------------------------------------------------------
//   Append the memory usage.
// -----------------------------------------------------------------------
unsigned int insert_mem_state(char *sbuff, unsigned int len) {
   FILE *f;
   int mem_total = -1;
   int mem_free = -1;
   int mem_cached = -1;
   int mem_buffers = -1;

   f = fopen(MEM_INFOS, "r");
   if (f == NULL) return 0;

   while (fgets(buff, BUFF_SZ, f)) {
      if (mem_total < 0 && sscanf(buff, "MemTotal:\t%d", &mem_total))
         continue;
      else if (mem_free < 0 && sscanf(buff, "MemFree:\t%d", &mem_free))
         continue;
      else if (mem_buffers < 0
            && sscanf(buff, "Buffers:\t%d", &mem_buffers))
         continue;
      else if (mem_cached < 0
            && sscanf(buff, "Cached:\t%d", &mem_cached))
         continue;
   }
   fclose(f);

   if (mem_total < 0 || mem_free < 0 || mem_cached < 0 || mem_buffers < 0)
      return 0;

   return snprintf(sbuff, len, "%d",
         (mem_total - mem_free - mem_cached - mem_buffers) / 1024);
}

// -----------------------------------------------------------------------
//   Append the swap usage.
// -----------------------------------------------------------------------
unsigned int insert_swap_state(char *sbuff, unsigned int len) {
   FILE *f;
   int swap_total = -1;
   int swap_free = -1;

   f = fopen(MEM_INFOS, "r");
   if (f == NULL) return 0;

   while (fgets(buff, BUFF_SZ, f)) {
      if (swap_total < 0 && sscanf(buff, "SwapTotal:\t%d", &swap_total))
         continue;
      else if (swap_free < 0 && sscanf(buff, "SwapFree:\t%d", &swap_free))
         continue;
   }
   fclose(f);

   if (swap_total < 0 || swap_free < 0)
      return 0;

   return snprintf(sbuff, len, "%d",
         (swap_total - swap_free) / 1024);
}

// -----------------------------------------------------------------------
//   Append the cpu frequency.
// -----------------------------------------------------------------------
unsigned int insert_cpu_freq(char *sbuff, unsigned int len,
      unsigned int number) {
   int cpu = -1;
   int curr_proc = -1;
   FILE *f;

   f = fopen(CPU_INFOS, "r");
   if (f == NULL) return 0;

   while (fgets(buff, BUFF_SZ, f)) {
      if (sscanf(buff, "processor\t: %d", &curr_proc))
         continue;
      else if (curr_proc == number && cpu < 0
            && sscanf(buff, "cpu MHz\t: %d", &cpu))
         continue;
   }
   fclose(f);

   if (cpu < 0) return 0;

   return snprintf(sbuff, len, "%d", cpu);
}

// -----------------------------------------------------------------------
//   Append the cpu frequency by cpufreq.
// -----------------------------------------------------------------------
void get_cpufreq_freq(unsigned int number, int *fcur, int *fpercent) {
   FILE *f;
   int fmax = -1;
   *fcur = 0;
   *fpercent = 0;

   snprintf(path, PATH_SZ, "%s%u%s", CPUFREQ_BASE, number, CPUFREQ_CUR);
   f = fopen(path, "r");

   if (f == NULL) return;

   fscanf(f, "%d", fcur);
   fclose(f);

   snprintf(path, PATH_SZ, "%s%u%s", CPUFREQ_BASE, number, CPUFREQ_MAX);
   f = fopen(path, "r");

   if (f == NULL) return;

   fscanf(f, "%d", &fmax);
   fclose(f);


   if (*fcur < 0 || fmax <= 0) return;

   *fpercent = (*fcur / (float) fmax) * 100;
   *fcur /= 1000;
}

// -----------------------------------------------------------------------
//   Append the temperature.
// -----------------------------------------------------------------------
unsigned int insert_temperature(char *sbuff, unsigned int len,
      const char *sensor_name) {
   int temp = -1;
   FILE *f;

   snprintf(path, PATH_SZ, "%s%s%s", THRM_ZONE, sensor_name, THRM_FILE);
   f = fopen(path, "r");

   if (f == NULL) return 0;

   while (fgets(buff, BUFF_SZ, f)) {
      if (temp < 0 && sscanf(buff, "temperature:\t%d", &temp))
         continue;
   }
   fclose(f);

   if (temp < 0) return 0;

   return snprintf(sbuff, len, "%d", temp);
}

// -----------------------------------------------------------------------
//   Append the network device.
// -----------------------------------------------------------------------
void get_network_device(char **devices, unsigned int nb_devices,
      char *name, unsigned int name_len, double *in, double *out) {
   FILE *f;
   char tmp_name[name_len];
   double tmp_in, tmp_out;

   for(int i = 0; i < nb_devices; i++) {
      unsigned long rx = 0, tx = 0, last_rx = 0, last_tx = 0;
      double timestamp = 0.0, last_timestamp = 0.0;
      int link_quality = 0;

      // Gets lasts rx/tx
      snprintf(path, PATH_SZ, "/var/tmp/skywmii_%s_data", devices[i]);

      f = fopen(path, "r");

      if (f != NULL) {
         while (fgets(buff, BUFF_SZ, f)) {
            if (sscanf(buff, "timestamp=%lf", &last_timestamp))
               continue;
            else if (sscanf(buff, "rx=%lu", &last_rx))
               continue;
            else if (sscanf(buff, "tx=%lu", &last_tx))
               continue;
         }

         fclose(f);
      }

      // Read download
      snprintf(path, PATH_SZ, "%s%s%s", NETWORK_FOLDER, devices[i],
            NETWORK_DOWN);

      f = fopen(path, "r");
      if (f == NULL) continue;
      
      if (!fscanf(f, "%lu", &rx)) {
         fclose(f);
         continue;
      }
      fclose(f);

      // Read upload
      snprintf(path, PATH_SZ, "%s%s%s", NETWORK_FOLDER, devices[i],
            NETWORK_UP);

      f = fopen(path, "r");
      if (f == NULL) continue;

      if (!fscanf(f, "%lu", &tx)) {
         fclose(f);
         continue;
      }
      fclose(f);

      timestamp = get_time();

      // Test if the device is a wireless device
      snprintf(path, PATH_SZ, "%s%s%s", WIRELESS_CHECK_BASE, devices[i],
            WIRELESS_CHECK_END);
      if (access(path, F_OK) == 0) {
         // Get link quality of the wireless device
         link_quality = get_wireless_quality(devices[i]);
      }

      if (((tx || rx) && (last_rx != rx || last_tx != tx))
            || i == 0) {
         // Compute transfers
         double rx_calc = 0.0, tx_calc = 0.0;

         rx_calc = compute_device_rate(timestamp, last_timestamp, rx,
               last_rx);

         tx_calc = compute_device_rate(timestamp, last_timestamp, tx,
               last_tx);

         // Writes result in status buffer
         if (link_quality) {
            snprintf(name, name_len, "%s[%d%%]", devices[i], link_quality);
         } else {
            strncpy(name, devices[i], name_len);
         }

         *in = rx_calc; *out = tx_calc;

         // Writes TX & RX in a temp file
         snprintf(path, PATH_SZ, "/var/tmp/skywmii_%s_data",
               devices[i]);

         f = fopen(path, "w");

         if (f != NULL) {
            fprintf(f, "timestamp=%lf\n", timestamp);
            fprintf(f, "rx=%lu\n", rx);
            fprintf(f, "tx=%lu\n", tx);

            fclose(f);
         }

         if (i > 0 || last_rx != rx || last_tx != tx) {
            break;
         }
      }
   }
}

// -----------------------------------------------------------------------
//   Compute the rate of a device.
// -----------------------------------------------------------------------
double compute_device_rate(double t_time, double t_last_time, unsigned
      long t_bytes, unsigned long t_last_bytes) {
   return ((t_bytes - t_last_bytes) / 1024.0) / (t_time - t_last_time);
}

// -----------------------------------------------------------------------
//   Get the link quality of a wireless device.
// -----------------------------------------------------------------------
int get_wireless_quality(const char *device) {
   FILE *f;
   unsigned int lq_max = 0, lq = 0;
   char sbuff [32];
   int skfd;
   struct iwreq iwr;
   struct iw_range range;

   // Read quality
   f = fopen(WIRELESS_INFOS, "r");

   if (f == NULL) return -1;

   snprintf(sbuff, 32, "%s:", device);
   while (fgets(buff, BUFF_SZ, f)) {
      if (strstr(buff, sbuff)) {
         if (sscanf(buff, "%*32[^:]:%*u%u.", &lq)) {

            // Now read the maximum quality
            if ((skfd = socket(AF_INET, SOCK_DGRAM, 0)) >= 0) {

               strncpy(iwr.ifr_ifrn.ifrn_name, device,
                     IFNAMSIZ);
               if (ioctl(skfd, SIOCGIWNAME, &iwr) >= 0) {

                  iwr.u.data.pointer = &range;
                  iwr.u.data.length = sizeof(struct iw_range);
                  iwr.u.data.flags = 0;
                  if (ioctl(skfd, SIOCGIWRANGE, &iwr) >= 0) 
                     lq_max = range.max_qual.qual;
               }
               close(skfd);
            }
         }
      }
   }
   fclose(f);

   if (lq_max <= 0) return 0;

   return (lq * 100) / lq_max;
}

#ifdef __WITH_XKB__
// -----------------------------------------------------------------------
//   Initialisation of XKB.
// -----------------------------------------------------------------------
int xkb_init(Display *dpy) {
   int xkb_opcode, xkb_event, xkb_error;
   int xkb_lmaj = XkbMajorVersion;
   int xkb_lmin = XkbMinorVersion;

   return XkbLibraryVersion(&xkb_lmaj, &xkb_lmin)
      && XkbQueryExtension(dpy, &xkb_opcode, &xkb_event, &xkb_error,
            &xkb_lmaj, &xkb_lmin);
}

// -----------------------------------------------------------------------
//   Get a mask modifier.
// -----------------------------------------------------------------------
unsigned int xkb_mask_modifier(XkbDescPtr xkb, const char *name) {
   unsigned int mask;
   if(!xkb || !xkb->names) return 0;

   for (int i = 0; i < XkbNumVirtualMods; i++) {
      char* modStr = XGetAtomName(xkb->dpy, xkb->names->vmods[i]);
      if (modStr != NULL && strcmp(name, modStr) == 0) {
         XkbVirtualModsToReal(xkb, 1 << i, &mask);
         return mask;
      }
   }
   return 0;
}

// -----------------------------------------------------------------------
//   Return the numlock mask.
// -----------------------------------------------------------------------
unsigned int xkb_numlock_mask(Display *dpy) {
   unsigned int mask;

   XkbDescPtr xkb;
   if ((xkb = XkbGetKeyboard(dpy, XkbAllComponentsMask, XkbUseCoreKbd))
         != NULL ) {
      mask = xkb_mask_modifier(xkb, "NumLock");
      XkbFreeKeyboard(xkb, 0, True);
      return mask;
   }
   return 0;
}

// -----------------------------------------------------------------------
//   Append the state of the numlock key.
// -----------------------------------------------------------------------
unsigned int insert_kbd_numlock(char *sbuff, unsigned int len) {
   unsigned int mask;
   unsigned int numlockState;

   // Open X display
   Display* dpy = XOpenDisplay( NULL );

   if (dpy == NULL) return 0;

   XkbStateRec xkbState;
   if (!xkb_init(dpy)) return 0;

   // Get mask of numlock
   mask = xkb_numlock_mask(dpy);
   if (mask == 0) return 0;

   // Get state of keybord
   XkbGetState(dpy, XkbUseCoreKbd, &xkbState);
   // Check if numlock is enabled or not
   numlockState = xkbState.locked_mods & mask;
   unsigned int rlen = 0;
   if (numlockState)
      rlen = snprintf(sbuff, len, "on");
   else
      rlen = snprintf(sbuff, len, "off");

   XCloseDisplay(dpy);

   return rlen;
}
#endif


// -----------------------------------------------------------------------
//   Append the temperature of a sensor.
// -----------------------------------------------------------------------
unsigned int insert_sensor_temp(char *sbuff, unsigned int len,
      unsigned int sensor_id, unsigned int temp_id) {
   FILE *f;
   int val = 0;

   snprintf(path, PATH_SZ, HWMON_THRM, sensor_id, temp_id);
   f = fopen(path, "r");
   if (f == NULL) return 0;

   fscanf(f, "%d", &val);
   fclose(f);

   return snprintf(sbuff, len, "%d", val / 1000);
}


// -----------------------------------------------------------------------
//   Append a fortune to the string
// -----------------------------------------------------------------------
unsigned int insert_fortune(char *sbuff, unsigned int len, const char* filename) {
   FILE* fortune = NULL;
   char filebuff[SBUFF_SZ];
   size_t bytes_read;
   char* tmp_char;

   if(filename) fortune = fopen(filename, "r");
   else fortune = fopen("/tmp/fortune", "r");

   if(!fortune) {
       fprintf(stderr, "Invalid fortune file: %s\n", filename);
       return 0;
   }
   bytes_read = fread(filebuff, 1, len, fortune);
   fclose(fortune);
   while((tmp_char = memchr(filebuff, '\n', bytes_read))) {
       *tmp_char = ' ';
   }
   memcpy(sbuff, filebuff, bytes_read);
   return bytes_read;
}
