// =======================================================================
// This file is part of sswriter (Sky Status Writer).
//
// Copyright 2008  Jonathan 'Sky' Squirawski <webmaster@sky-siteweb.com>
// 
// SSWriter is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// SSWriter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SSWriter.  If not, see <http://www.gnu.org/licenses/>.
// =======================================================================

#include "cm_parsing.h"
#include "sswriter_std.h"
#include "acpi_parsing.h"
#include <string.h>

// -----------------------------------------------------------------------
//   Print the help usage of skywmii.
// -----------------------------------------------------------------------
void print_usage(const char *progname) {
   printf("Usage: %s [options] <format>\n\n", progname);
   printf("format: a format string for the output.\n\n");
   printf("%%b<n>\tcharge of battery number <n>\n");
   printf("%%c<n>\tcpu number <n> frequency (using /proc/cpuinfo)\n");
   printf("%%C<n>\tcpu number <n> frequency (using cpufreq)\n");
   printf("%%d\tdate\n");
   printf("%%i\tin rate of the network device\n");
   printf("%%l\tstate of the numlock key\n");
   printf("%%m\tmemory usage\n");
   printf("%%n\tname of the network device\n");
   printf("%%o\tout rate of the network device\n");
   printf("%%p<n>\tcpu number <n> frequency percent (using cpufreq)\n");
   printf("%%s\tswap usage\n");
   printf("%%f\tfortune\n");
   printf("%%t<n>\ttemperature of the sensor number <n> specified by -t\n");
   printf("%%T<n>\ttemperature of the sensor number <n> specified by -T\n");
   printf("%%%%\ta single %%\n");
   printf("\nOptions:\n");
   printf("-h --help\t\t\tprint help\n");
   printf("-d --date\t\t\tformat string of the date and time (see strftime)\n");
   printf("-w --warnlimit\t<number>\tset the limit of the battery warning\n");
   printf("-D --device\t<device1 ... deviceN>\n");
   printf("\t\t\t\tadd in order the devices\n");
   printf("\t\t\t\tex: %s -D eth0 ra0 \"Network: %%n\"\n", progname);
   printf("-t --tsensors\t<sensor1 ... sensorN>\n");
   printf("\t\t\t\tspecify the name of the temperatures sensors\n");
   printf("-T --lm-tsensors\t<hwmon-num,temp-num ...>\n");
   printf("\t\t\t\tspecify differents couples of hwmon-num/temp-num (see README)\n");
   printf("-f --fortune\tset the file with a fortune in it. (see README for wmii integration)\n");
}


// -----------------------------------------------------------------------
//   Parsing of arguments.
// -----------------------------------------------------------------------
int parse_args(int argc, char **argv) {
   unsigned int nb_devices = 0;
   char **devices = NULL;
   unsigned int battery_warn_limit = 10;
   char *date_format = NULL;
   char *fortune_file = NULL;
   char **sensors = NULL;
   unsigned int nb_sensors = 0;
   unsigned int nb_lm_sensors = 0;
   unsigned int *lm_sensors = NULL;

   int i = 0;
   for (i = 1; i < argc; i++) {
      if (!strcmp(argv[i], "-h")
            || !strcmp(argv[i], "--help")) {
         print_usage(argv[0]);
         return EXIT_SUCCESS;
      } else if (!strcmp(argv[i], "-w") || !strcmp(argv[i],
               "--warnlimit")) {
         i++;
         if (i >= argc) break;
         sscanf(argv[i], "%u", &battery_warn_limit);
      } else if (!strcmp(argv[i], "-d") || !strcmp(argv[i], "--date")) {
         i++;
         if (i >= argc) break;
         date_format = argv[i];
      } else if (!strcmp(argv[i], "-f") || !strcmp(argv[i], "--fortune")) {
         i++;
         if (i >= argc) break;
         fortune_file = argv[i];
      } else if (devices == NULL && (!strcmp(argv[i], "-D")
               || !strcmp(argv[i], "--device"))) {
         int j = i + 1;
         while (j < (argc - 1) && argv[j][0] != '-') {
            nb_devices++;
            j++;
         }

         if (nb_devices > 0) {
            devices = calloc(nb_devices, sizeof(char*));
            for (j = 0; j < nb_devices; j++) {
               devices[j] = calloc(16, sizeof(char));
               strncpy(devices[j], argv[i + j + 1], 16);
            }
            i += j;
         }
      } else if (sensors == NULL && (!strcmp(argv[i], "-t")
               || !strcmp(argv[i], "--tsensors"))) {
         int j = i + 1;
         while (j < (argc - 1) && argv[j][0] != '-') {
            nb_sensors++;
            j++;
         }

         if (nb_sensors > 0) {
            sensors = calloc(nb_sensors, sizeof(char*));
            for (j = 0; j < nb_sensors; j++) {
               sensors[j] = calloc(16, sizeof(char));
               strncpy(sensors[j], argv[i + j + 1], 16);
            }
            i += j;
         }
      } else if (lm_sensors == NULL && (!strcmp(argv[i], "-T")
               || !strcmp(argv[i], "--lm-tsensors"))) {
         int j = i + 1;
         while (j < (argc - 1) && argv[j][0] != '-') {
            nb_lm_sensors++;
            j++;
         }

         if (nb_lm_sensors > 0) {
            lm_sensors = calloc(2 * nb_lm_sensors, sizeof(unsigned int));
            for (j = 0; j < nb_lm_sensors; j++) {
               unsigned int v1, v2;
               if (sscanf(argv[i + j + 1], "%u,%u", &v1, &v2) == 2) {
                  lm_sensors[j * 2] = v1;
                  lm_sensors[j * 2 + 1] = v2;
               } else {
                  print_usage(argv[0]);
                  return EXIT_FAILURE;
               }
            }
            i += j;
         }

      } else break;
   }

   int res = 0;

   if (i == argc - 1) {
      res = handle_format(argv[i], devices, nb_devices,
            battery_warn_limit, date_format, fortune_file, sensors, 
            nb_sensors, lm_sensors, nb_lm_sensors);
   } else {
      print_usage(argv[0]);
   }

   // Free memory
   if (nb_devices > 0) {
      for (i = 0; i < nb_devices; i++) {
         free(devices[i]);
      }
      free(devices);
   }
   if (nb_sensors > 0) {
      for (i = 0; i < nb_sensors; i++) {
         free(sensors[i]);
      }
      free(sensors);
   }
   if (nb_lm_sensors > 0) {
      free(lm_sensors);
   }

   return res;
}

// -----------------------------------------------------------------------
//   Handling of the formated string.
// -----------------------------------------------------------------------
int handle_format(const char *format, char **net_devices,
      unsigned int net_nb_devices, unsigned int battery_warn_limit,
      const char *date_format, const char *fortune_file, char **sensors, 
      unsigned int nb_sensors, unsigned int *lm_sensors, unsigned int nb_lm_sensors) {
   char *it = NULL;

   char line[SBUFF_SZ];
   char *line_it = line;
   unsigned int line_left = SBUFF_SZ - 1;
   unsigned int len = 0;
   char cnum = 0;

   // Network device cache
   int net_cached = 0;
   char net_name[32] = "";
   double net_in = 0.0, net_out = 0.0;

   // Cpufreq cache
   int fcached = -1;
   int ffreq = 0;
   int fpercent = 0;

   int warn = 0;
   
   memset(line, '\0', SBUFF_SZ);

   // Looping though format string
   for (it = (char*) format; ((*it) != '\0') && (line_left > 0); it++) {
      char c = (*it);
      char next = 0;

      if (c == '%') {
         // It's a control sequence
         it++;
         next = (*it);

         switch (next) {
            case '%':
               (*line_it) = '%';
               len = 1;
            break;
            case 'd': // Date
               len = insert_date(line_it, line_left, date_format);
            break;
            case 'b': // Battery
               it++;
               cnum = (*it);
               if (isdigit(cnum)) {
                  len = insert_battery_state(line_it, line_left,
                        ctoi(cnum), battery_warn_limit, &warn);
                  if (warn) {
                     printf("%s", line_it);
                     return EXIT_SUCCESS;
                  }
               } else {
                  fprintf(stderr, "A number was expected after %%b.\n");
                  return EXIT_FAILURE;
               }
            break;
            case 'm': // Memory
               len = insert_mem_state(line_it, line_left);
            break;
            case 's': // Swap
               len = insert_swap_state(line_it, line_left);
            break;
            case 'c': // Cpu
               it++;
               cnum = (*it);
               if (isdigit(cnum)) {
                  len = insert_cpu_freq(line_it, line_left, ctoi(cnum));
               } else {
                  fprintf(stderr, "A number was expected after %%c.\n");
                  return EXIT_FAILURE;
               }
            break;
            case 'C': // Cpu by cpufreq
               it++;
               cnum = (*it);
               if (isdigit(cnum)) {
                  int num = ctoi(cnum);

                  if (fcached != num) {
                     fcached = num;
                     get_cpufreq_freq(num, &ffreq, &fpercent);
                  }

                  len = snprintf(line_it, line_left, "%d", ffreq);
               } else {
                  fprintf(stderr, "A number was expected after %%C.\n");
                  return EXIT_FAILURE;
               }
            break;
            case 'p': // Cpu frequency percent by cpufreq
               it++;
               cnum = (*it);
               if (isdigit(cnum)) {
                  int num = ctoi(cnum);

                  if (fcached != num) {
                     fcached = num;
                     get_cpufreq_freq(num, &ffreq, &fpercent);
                  }

                  len = snprintf(line_it, line_left, "%d", fpercent);
               } else {
                  fprintf(stderr, "A number was expected after %%p.\n");
                  return EXIT_FAILURE;
               }
            break;

            case 't': // Thermal zone
               it++;
               cnum = (*it);
               if (isdigit(cnum)) {
                  int num = ctoi(cnum);
                  if (num < nb_sensors) {
                     len = insert_temperature(line_it, line_left,
                           sensors[num]);
                  } else {
                     fprintf(stderr,
                           "The sensor %d isn't specified by -t\n", num);
                     return EXIT_FAILURE;
                  }
               } else {
                  fprintf(stderr, "A number was expected after %%t.\n");
                  return EXIT_FAILURE;
               }
            break;
            case 'T': // Thermal hwmon sensor
               it++;
               cnum = (*it);
               if (isdigit(cnum)) {
                  int num = ctoi(cnum);
                  if (num * 2 < nb_lm_sensors) {
                     num *= 2;
                     len = insert_sensor_temp(line_it, line_left,
                           lm_sensors[num], lm_sensors[num + 1]);
                  } else {
                     fprintf(stderr,
                           "The sensor %d isn't specified by -T\n", num);
                     return EXIT_FAILURE;
                  }
               } else {
                  fprintf(stderr, "A number was expected after %%T.\n");
                  return EXIT_FAILURE;
               }
            break;


            case 'n': // Network devices -- name
               if (!net_cached) {
                  get_network_device(net_devices, net_nb_devices,
                        net_name, 32, &net_in, &net_out);
                  net_cached = 1;
               }

               len = snprintf(line_it, line_left, "%s", net_name);
            break;
            case 'i': // Network devices -- in
               if (!net_cached) {
                  get_network_device(net_devices, net_nb_devices,
                        net_name, 32, &net_in, &net_out);
                  net_cached = 1;
               }

               len = snprintf(line_it, line_left, "%.1lf", net_in);
            break;
            case 'o': // Network devices -- out
               if (!net_cached) {
                  get_network_device(net_devices, net_nb_devices,
                        net_name, 32, &net_in, &net_out);
                  net_cached = 1;
               }

               len = snprintf(line_it, line_left, "%.1lf", net_out);
            break;

            case 'l':
#ifdef __WITH_XKB__
               len = insert_kbd_numlock(line_it, line_left);
#else
               fprintf(stderr,
                     "Sswriter needs the support of XKB to use this"
                     " feature.\n");
               return EXIT_FAILURE;
#endif
            break;
            
            case 'f': // Fortune
               len = insert_fortune(line_it, line_left, fortune_file);
            break;

            default:
               fprintf(stderr, "Invalid format sequence: %%%c.\n", next);
               return EXIT_FAILURE;
         }

         line_it += len;
         line_left -= len;
      } else if (c == '\\') {
         // Handling escape sequence
         it++;
         next = (*it);

         switch (next) {
            case '\\':
               (*line_it) = '\\';
            break;
            case 'n':
               (*line_it) = '\n';
            break;
            case 't':
               (*line_it) = '\t';
            break;
            case 'r':
               (*line_it) = '\r';
            break;
            default:
               fprintf(stderr, "Invalid escape sequence: \\%c.\n", next);
               return EXIT_FAILURE;
         }
         line_it++;
         line_left--;
      } else {
         // Simple char
         (*line_it) = c;
         line_it++;
         line_left--;
      }
   }
   
   printf("%s", line);
   return EXIT_SUCCESS;
}
